# -*- coding: utf-8 -*-
"""
Propulsions Model Restructuring (8/29/2025)
File structure:
    - Numerical methods for numba calculation (bisection, secant, brent, etc)
    - Preprocessing functions for propeller, motor, battery, ESC data
    - PointResult function
    - LinePlot functions (one for each input combo)
    - ContourPlot functions
    - CubicPlot functions
    - Pareto Front functions
    - Mission model compatible functions
    - GEKKO prop model functions for reference 

Primary inputs:
    Battery State Of Charge (SOC) as %
    Freestream Velocity (Vinf) as m/s
    Throttle Setting (dT) as %
    
LinePlot function:
    By fixing two of these inputs, a line plot of a propulsion quantity (propQ) can be plotted with respect to the third.

ContourPlot function:
    By fixing one of these inputs, a contour plot of a propQ is plotted wrt the two unfixed inputs.

CubicPlot function:
    Additionally a cubic plot of propQ variation with all three variables is available, but usually hard to use in reports.

for all of these functions, SOC can be provided as SOC (%), Voc (Volt), or t (s), which assumes constant current (i.e. good for aircraft in cruise)

Available propQs are:
    T (lbf)     (thrust for all motors)
    Q (N*m)     (torque for all motors)
    RPM         (for a single motor/propeler)

nondimensional:
    CT          (propeller (prop) torque constant)
    CP          (propeller power constant)
    eta_p       (propeller efficiency)
    eta_m       (motor efficiency)
    eta_c       (controller (ESC) efficiency)
    eta_drive   (drive efficiency)

all in Watts (W):
    Pout    (mechanical W for a single motor)
    Pin_m   (electric W input to a single motor (equivalent to Pout from a single ESC))
    Pin_c   (electric W input to all ESCs (equivalent to Pout from the battery))
    
all in Ampere (A):
    Im      (motor current for a single motor)
    Ic      (ESC current for a single ESC)
    Ib      (battery current)
    
all in Volts (V):
    Voc     (cell voltage)
    Vb      (battery voltage)
    Vm      (motor voltage)
    Vc      (ESC voltage)
    
    
Propeller coefficients (CT, CP) are primarily acquired by interpolation between the APC technical datasheets 
(see: https://www.apcprop.com/technical-information/performance-data/?v=7516fd43adaa)

These datasheets were generated using Blade Element Theory in conjunction with NASA TAIR and some databases for airfoil data. 
They are NOT fully accurate, especially for propeller stall, which mostly occurs when propellers low diam/pitch ratios travel at lower freestream velocities. 

Use the APCBEMTvsUIUCexpdata.py file to automatically compare APC electric propeller data from both sources!

While a wonderful amount of experimental data has been gathered by Michael Selig of UIUC's group (see: https://m-selig.ae.illinois.edu/props/propDB.html),
this data does not extend into the high performance ranges (say RPMs of 6k-12k for ~15-20 inch diameter propellers) that typically occur with 6-12S liPo batteries.

As of 8/29/2025, this experimental data is not implemented yet, but in the future it will be used in conjunction with the APC data for a mixed fidelity approach.


Model Formulation (primary based on Saemi 2023, secondary if available data based on Gong 2018):
    Simplified RPM (with constant ESC efficiency and constant I0)
    SAEMI 2023: https://www.mdpi.com/2226-4310/11/1/16
    GONG 2018:  https://www.researchgate.net/profile/Andrew-Gong-2/publication/326263042
    Jeong 2020: https://www.researchgate.net/publication/347270768_Improvement_of_Electric_Propulsion_System_Model_for_Performance_Analysis_of_Large-Size_Multicopter_UAVs
    
LiPo discharge curve (Voc(SOC)) from Chen 2006 (https://rincon-mora.gatech.edu/publicat/jrnls/tec05_batt_mdl.pdf):
    Voc = 3.685 - 1.031 * np.exp(-35 * SOC) + 0.2156 * SOC - 0.1178 * SOC**2 + 0.3201 * SOC**3
    
    Alternative from Jeong 2020
    Voc = 1.7*(SOC**3) - 2.1*(SOC**2) + 1.2*SOC + 3.4

In reality the discharge curve (and total capacity) is strongly influenced by cell temperature and battery health. 
Corrections on Voc for battery health are optionally defined by inputting Voc at maximum charge (determined experimentally) (NOT IMPLEMENTED AS OF 8/29/2025).
    
All of the following constants used in the propulsion model are aquired via the Motors, Batteries, and ESCs CSV sheets, which can all be adjusted by any user.
Motor constants:    KV (RPM/V), I0 (A), Rm (Ohm)
Battery constants:  CB (mAh), ns (number of cells in series), Rb (Ohm)

ESC constants (Saemi):      Rds (Ohm), fPWM (Hz), Tsd (s), Psb (W)   <---- currently defaults from Saemi 2023 are used
ESC constants (Gong):       a_m (the slope of constant a), a_0 (y intercept), b, c_m, c_0  <---- only available for a very limited selection of tested ESCs
ESC constnats (Jeong):       #TODO FILL IN JEONG CONSTANTS
    
KV = speed constant, I0 = no-load current, Rm = motor resistance
CB = battery capacity, ns = number of cells (in series), Rb = battery resistance

Simplifed RPM formulation (known Vsoc, Vinf, dT):
    
    RPM guess
    J = Vinf/((RPM/60)*d)
    CP = CPNumba(RPM, J, rpm_list, coef_numba_prop_data)
    Q = rho*((RPM/60)**2)*(d**5)*(CP/(2*np.pi))
    Im = Q*KV*(np.pi/30) + I0 # for one motor
    Ib = (nmot/eta_c)*Im
    Vb = ns*(Voc) - Ib*Rb
    Vm = dT*Vb
    RPMcalc = KV*(Vm - Im*Rm)
    res = RPMcalc - RPM

Simplified RPM formulation (known runtime (t), constant current):

    RPM guess
    J = Vinf/((RPM/60)*d)
    CP = CPNumba(RPM, J, rpm_list, coef_numba_prop_data)
    Q = rho*((RPM/60)**2)*(d**5)*(CP/(2*np.pi))
    Im = Q*KV*(np.pi/30) + I0 # for one motor
    Ib = (nmot/eta_c)*Im
    SOC = 1.0 - (Ib*t)/(CB*3.6)
    Voc = 3.685 - 1.031 * np.exp(-35 * SOC) + 0.2156 * SOC - 0.1178 * SOC**2 + 0.3201 * SOC**3
    Vb = ns*(Voc) - Ib*Rb
    Vm = dT*Vb
    RPMcalc = KV*(Vm - Im*Rm)
    res = RPMcalc - RPM
    
minimize res!

This formulation is advantageous for its computational efficiency. With only one variable, plots can be made extremely quickly.
Significant loss of accuracy due to the constant I0 assumption compared to the other models.

    
@author: NASSAS
"""

import numpy as np
from numpy.polynomial import Polynomial
import matplotlib.pyplot as plt
from tqdm import tqdm
from numba import njit
import copy
from uavdex.VSPcontribution.atmosphere import stdatm1976 as atm 


lbfN = 4.44822
ftm = 0.3048
MPH_TO_MPS = 0.44704  # Conversion factor: 1 mph to m/s

global propQnames, propQshort
propQnames = ['Total Thrust (N)', 
              'Total Torque (Nm)', 
              'RPM', 
              'Drive Efficiency', 
              'Propeller Efficiency', 
              'Gearing Efficiency', 
              'Motor Efficiency', 
              'ESC Efficiency', 
              'Battery Efficiency', 
              'Mech. Power Out of 1 Motor (W)', 
              'Elec. Power Into 1 Motor (W)', 
              'Elec. Power Into 1 ESC (W)', 
              'Waste Power in 1 Motor (W)', 
              'Waste Power in 1 ESC (W)',
              'Waste Power in 1 Battery (W)',
              'Current in 1 Motor (A)', 
              'Current in 1 ESC (A)', 
              'Current in Battery (A)',
              'Voltage in 1 Motor (V)', 
              'Voltage in 1 ESC (V)', 
              'Battery Voltage (V)', 
              'Voltage Per Cell (V)', 
              'State of Charge']
propQshort = ['T', 'Q', 'RPM', 
              'eta_drive', 'eta_p', 'eta_g', 'eta_m', 'eta_c', 'eta_b', 
              'Pout', 'Pin_m', 'Pin_c', 'Pw_m', 'Pw_c', 'Pw_b', 
              'Im', 'Ic', 'Ib', 'Vm', 'Vc', 'Vb', 
              'Voc', 'SOC']

from uavdex import _uavdex_root
path_to_data = _uavdex_root / 'Databases/'

### OLD LOCAL TESTING
# from VSPcontribution.atmosphere import stdatm1976 as atm
# from pathlib import Path
# _uavdex_root = Path(__file__).parent
# path_to_data = _uavdex_root / 'Databases/'

#%%################### Numerical Methods ###################
# 1D root finding 
@njit(fastmath = True)
def bisection(low, high, func, *args):
    tol = 1e-3
    max_iter = 100
    for _ in range(max_iter):
        mid = (low + high) / 2
        res_mid = func(mid, *args)
        
        if res_mid**2 < tol**2 or res_mid == 0.0:
            return(mid)
        
        if res_mid*func(low, *args) < 0:
            high = mid
        else:
            low = mid
    else:
        # essentially error message
        return(-1.0)

@njit(fastmath = True)
def secant(low, high, func, *args):
    tol = 1e-3
    max_iter = 100 
    
    x0 = low 
    x1 = high
    for i in range(max_iter):
        x2 = x1 - func(x1, *args) * (x1 - x0) / (func(x1, *args) - func(x0, *args))
        x0, x1 = x1, x2
        if (x0 - x1)**2 < tol**2 or func(x1, *args)**2 < tol**2:
            return(x2)
    else:
        return(-1.0)

# non-numba version for faster single point analysis
def bisectionBase(low, high, func, *args):
    tol = 1e-3
    max_iter = 100
    for _ in range(max_iter):
        mid = (low + high) / 2
        res_mid = func(mid, *args)
        
        if res_mid**2 < tol**2 or res_mid == 0.0:
            return(mid)
        
        if res_mid*func(low, *args) < 0:
            high = mid
        else:
            low = mid
    else:
        # essentially error message
        return(-1.0)

#%% ################### Data Parsing ###################
def parse_coef_propeller_data(prop_name):
    """
    prop_name in the form: 16x10E, 18x12E, 12x12, etc (no PER3_ and no .dat to make it easier for new users)
    Parses the provided PER3_16x10E.dat content to extract RPM, V (m/s), Thrust (N), Torque (N-m).
    Stores in PROP_DATA as {rpm: {'V': np.array, 'Thrust': np.array, 'Torque': np.array}}
    """    
    PROP_DATA = {}

    with open(path_to_data / f'APCPropDatabase/PER3_{prop_name}.dat', 'r') as f:
        data_content = f.read()

    current_rpm = None
    in_table = False
    table_lines = []
    
    for line in data_content.splitlines():
        line = line.strip()
        if line.startswith("PROP RPM ="):
            # Extract RPM
            current_rpm = int(line.split("=")[-1].strip())
            in_table = False
            table_lines = []
        elif line.startswith("V") and "J" in line and current_rpm is not None:
            # Start of table headers
            in_table = True
        elif in_table and line and not line.startswith("(") and len(line.split()) >= 10:
            # Parse data rows (ensure it's a data line with enough columns)
            parts = line.split()
            try:
                J = float(parts[1])  # advance ratio J
                CT = float(parts[3])  # thrust coef
                CP = float(parts[4])  # power coef (can convert to CQ)
                # v_mps = v_mph * MPH_TO_MPS  # Convert to m/s
                table_lines.append((J, CT, CP))
            except (ValueError, IndexError):
                continue  # Skip malformed lines
        elif in_table and (line == "" or "PROP RPM" in line):
            # End of table for this RPM, store if data exists
            if current_rpm and table_lines:
                J_list, CT_list, CP_list = zip(*sorted(table_lines))  # Sort by V for interp1d
                PROP_DATA[current_rpm] = {
                    'J': np.array(J_list),
                    'CT': np.array(CT_list),
                    'CP': np.array(CP_list)
                }
            in_table = False
    
    # Sort RPM keys for efficient lookup
    PROP_DATA['rpm_list'] = sorted(PROP_DATA.keys())
    
    # array based datastructure where each index corresponds to rpm_values[i] (or i+1*1000 RPM)
    # and in each index there is [[V values], [Thrust values], [Torque values]] at the indices, 0, 1, 2
    numba_prop_data = []
    for RPM in PROP_DATA['rpm_list']:
        datasection = np.array([PROP_DATA[RPM]['J'], 
                                PROP_DATA[RPM]['CT'], 
                                PROP_DATA[RPM]['CP']])
        numba_prop_data.append(datasection)
        
    return(PROP_DATA, numba_prop_data)

def parse_propeller_data(prop_name):
    """
    prop_name in the form: 16x10E, 18x12E, 12x12, etc (no PER3_ and no .dat to make it easier for new users)
    Parses the provided PER3_16x10E.dat content to extract RPM, V (m/s), Thrust (N), Torque (N-m).
    Stores in PROP_DATA as {rpm: {'V': np.array, 'Thrust': np.array, 'Torque': np.array}}
    """    
    PROP_DATA = {}

    with open(path_to_data / f'APCPropDatabase/PER3_{prop_name}.dat', 'r') as f:
        data_content = f.read()

    current_rpm = None
    in_table = False
    table_lines = []
    
    for line in data_content.splitlines():
        line = line.strip()
        if line.startswith("PROP RPM ="):
            # Extract RPM
            current_rpm = int(line.split("=")[-1].strip())
            in_table = False
            table_lines = []
        elif line.startswith("V") and "J" in line and current_rpm is not None:
            # Start of table headers
            in_table = True
        elif in_table and line and not line.startswith("(") and len(line.split()) >= 10:
            # Parse data rows (ensure it's a data line with enough columns)
            parts = line.split()
            try:
                v_mph = float(parts[0])  # V in mph
                torque_nm = float(parts[9])  # Torque (N-m)
                thrust_n = float(parts[10])  # Thrust (N)
                v_mps = v_mph * MPH_TO_MPS  # Convert to m/s
                table_lines.append((v_mps, thrust_n, torque_nm))
            except (ValueError, IndexError):
                continue  # Skip malformed lines
        elif in_table and (line == "" or "PROP RPM" in line):
            # End of table for this RPM, store if data exists
            if current_rpm and table_lines:
                v_list, thrust_list, torque_list = zip(*sorted(table_lines))  # Sort by V for interp1d
                PROP_DATA[current_rpm] = {
                    'V': np.array(v_list),
                    'Thrust': np.array(thrust_list),
                    'Torque': np.array(torque_list)
                }
            in_table = False
    
    # Sort RPM keys for efficient lookup
    PROP_DATA['rpm_list'] = sorted(PROP_DATA.keys())
    
    # array based datastructure where each index corresponds to rpm_values[i] (or i+1*1000 RPM)
    # and in each index there is [[V values], [Thrust values], [Torque values]] at the indices, 0, 1, 2
    numba_prop_data = []
    for RPM in PROP_DATA['rpm_list']:
        datasection = np.array([PROP_DATA[RPM]['V'], 
                                PROP_DATA[RPM]['Thrust'], 
                                PROP_DATA[RPM]['Torque']])
        numba_prop_data.append(datasection)
        
    return(PROP_DATA, numba_prop_data)

def initialize_RPM_polynomials(PROP_DATA):
    """
    returns: rpm_values, thrust_polys, torque_polys, V_DOMAINS
    
    Creates polynomial approximations for thrust and torque that are compatible with GEKKO in the form of Thrust(V) for a fixed RPM
    Uses piecewise polynomials for different RPM ranges.
    """
    
    # Extract data for polynomial fitting
    rpm_values = sorted([rpm for rpm in PROP_DATA.keys() if isinstance(rpm, int)])
    
    # Create coefficient matrices for polynomial approximation
    # We'll use separate polynomials for different velocity ranges
    thrust_polys = {}
    torque_polys = {}
    
    V_DOMAINS = []
    for rpm in rpm_values:
        data = PROP_DATA[rpm]
        V_data = data['V']
        thrust_data = data['Thrust']
        torque_data = data['Torque']
                
        # Fit polynomials (degree 3-4 should be sufficient for most cases)
        thrust_poly = Polynomial.fit(V_data, thrust_data, deg=4)
        thrust_polys[rpm] = thrust_poly
        
        # Fit polynomial to torque data  
        torque_poly = Polynomial.fit(V_data, torque_data, deg=4)
        torque_polys[rpm] = torque_poly
    
        V_DOMAINS.append(torque_poly.domain[1])
    
    V_DOMAINS = np.array(V_DOMAINS)
    return rpm_values, thrust_polys, torque_polys, V_DOMAINS

#%% Numba Interpolation Functions for J, CT, CP data
@njit(fastmath = True)
def CPNumba(RPM, J, rpm_list, numba_prop_data):
    '''
    J: advance ratio

    numba_prop_data is packaged so each index corresponds to (i+1)*1000 RPM 
    with the structure [[Jvalues], [CTvalues], [CPvalues]] for each index
    data[0] = J values, data[1] = CT, data[2] = CP
    '''
    if RPM < rpm_list[0] or RPM > rpm_list[-1] or J < 0:
        return 0.0
    
    idx = np.searchsorted(rpm_list, RPM)
    if idx == 0:
        closest_rpms = [rpm_list[0]]
    elif idx == len(rpm_list):
        closest_rpms = [rpm_list[-1]]
    else:
        closest_rpms = [rpm_list[idx - 1], rpm_list[idx]]
        
    CPs = []
    for rpm in closest_rpms:
        data = numba_prop_data[int(rpm/1000 -1)]
        # NEW CODE WAS NEEDED TO BRING IT TO 0 WHEN J WAS OUTSIDE BOUNDS!!
        if J > data[0].max():
            CPs.append(0.0)
            continue
        CPs.append(np.interp(J, data[0], data[2]))
        
    CPs = np.array(CPs)
    
    if len(closest_rpms) == 1:
        return CPs[0]
    else:
        weight = (RPM - closest_rpms[0]) / (closest_rpms[1] - closest_rpms[0])
        return (1 - weight) * CPs[0] + weight * CPs[1]

@njit(fastmath = True)
def CTNumba(RPM, J, rpm_list, numba_prop_data):
    '''
    J: advance ratio
    
    numba_prop_data is packaged so each index corresponds to (i+1)*1000 RPM 
    with the structure [[Jvalues], [CTvalues], [CPvalues]] for each index
    data[0] = J values, data[1] = CT, data[2] = CP
    '''
    if RPM < rpm_list[0] or RPM > rpm_list[-1] or J < 0:
        return 0.0
    
    idx = np.searchsorted(rpm_list, RPM)
    if idx == 0:
        closest_rpms = [rpm_list[0]]
    elif idx == len(rpm_list):
        closest_rpms = [rpm_list[-1]]
    else:
        closest_rpms = [rpm_list[idx - 1], rpm_list[idx]]
        
    CTs = []
    for rpm in closest_rpms:
        data = numba_prop_data[int(rpm/1000 -1)]
        if J > data[0].max():
            CTs.append(0.0)
            continue
        CTs.append(np.interp(J, data[0], data[1]))
        
    CTs = np.array(CTs)
    
    if len(closest_rpms) == 1:
        return CTs[0]
    else:
        weight = (RPM - closest_rpms[0]) / (closest_rpms[1] - closest_rpms[0])
        return (1 - weight) * CTs[0] + weight * CTs[1]
    
#%% Non-numba interpolation functions
def CPBase(RPM, J, rpm_list, numba_prop_data):
    '''
    J: advance ratio

    numba_prop_data is packaged so each index corresponds to (i+1)*1000 RPM 
    with the structure [[Jvalues], [CTvalues], [CPvalues]] for each index
    data[0] = J values, data[1] = CT, data[2] = CP
    '''
    if RPM < rpm_list[0] or RPM > rpm_list[-1] or J < 0:
        return 0.0
    
    idx = np.searchsorted(rpm_list, RPM)
    if idx == 0:
        closest_rpms = [rpm_list[0]]
    elif idx == len(rpm_list):
        closest_rpms = [rpm_list[-1]]
    else:
        closest_rpms = [rpm_list[idx - 1], rpm_list[idx]]
        
    CPs = []
    for rpm in closest_rpms:
        data = numba_prop_data[int(rpm/1000 -1)]
        # NEW CODE WAS NEEDED TO BRING IT TO 0 WHEN J WAS OUTSIDE BOUNDS!!
        if J > data[0].max():
            CPs.append(0.0)
            continue
        CPs.append(np.interp(J, data[0], data[2]))
        
    CPs = np.array(CPs)
    
    if len(closest_rpms) == 1:
        return CPs[0]
    else:
        weight = (RPM - closest_rpms[0]) / (closest_rpms[1] - closest_rpms[0])
        return (1 - weight) * CPs[0] + weight * CPs[1]

def CTBase(RPM, J, rpm_list, numba_prop_data):
    '''
    J: advance ratio
    
    numba_prop_data is packaged so each index corresponds to (i+1)*1000 RPM 
    with the structure [[Jvalues], [CTvalues], [CPvalues]] for each index
    data[0] = J values, data[1] = CT, data[2] = CP
    
    '''
    if RPM < rpm_list[0] or RPM > rpm_list[-1] or J < 0:
        return 0.0
    
    idx = np.searchsorted(rpm_list, RPM)
    if idx == 0:
        closest_rpms = [rpm_list[0]]
    elif idx == len(rpm_list):
        closest_rpms = [rpm_list[-1]]
    else:
        closest_rpms = [rpm_list[idx - 1], rpm_list[idx]]
    
    CTs = []
    for rpm in closest_rpms:
        data = numba_prop_data[int(rpm/1000 - 1)]
        
        # NEW CODE WAS NEEDED TO BRING IT TO 0 WHEN J WAS OUTSIDE BOUNDS!!
        if J > data[0].max():
            CTs.append(0.0)
            continue
        
        CTs.append(np.interp(J, data[0], data[1]))
    CTs = np.array(CTs)
    
    if len(closest_rpms) == 1:
        return CTs[0]
    else:
        weight = (RPM - closest_rpms[0]) / (closest_rpms[1] - closest_rpms[0])
        return (1 - weight) * CTs[0] + weight * CTs[1]


#%% ####################### PROPULSION MODELS #######################
@njit(fastmath = True)
def VocFunc(SOC, BattType):
    '''
    Determines the battery Voltage (Voc) as a function of State of Charge (SOC) given from 0-1 for a specified battery chemistry
    
    LiPo
    Main equation from Chen 2006
    https://rincon-mora.gatech.edu/publicat/jrnls/tec05_batt_mdl.pdf
    
    Alternative from Jeong 2020
    Voc = 1.7*(SOC**3) - 2.1*(SOC**2) + 1.2*SOC + 3.4
    https://www.researchgate.net/publication/347270768_Improvement_of_Electric_Propulsion_System_Model_for_Performance_Analysis_of_Large-Size_Multicopter_UAVs
    
    NOTE:
    Jeong 2020 also presents resistance as a function of cell energy:
    Rbatt.cell = 21.0*(ebatt.cell)**-0.8056
    
    
    Liion
    https://www.researchgate.net/publication/346515863_Comparison_of_Lithium-Ion_Battery_Pack_Models_Based_on_Test_Data_from_Idaho_and_Argonne_National_Laboratories
    Using the i3 battery pack as an intermediate value
    
    TODO: implment battery voltage equation adjustments based on predicted health (measured via Vsoc at full charge!)
    '''
    if BattType == 'LiPo':
        return(3.685 - 1.031 * np.exp(-35 * SOC) + 0.2156 * SOC - 0.1178 * SOC**2 + 0.3201 * SOC**3)
    elif BattType == 'Liion':
        return(-4.48*((1-SOC)**5) + 9.09*((1-SOC)**4) - 7.08*((1-SOC)**3) + 2.32*((1-SOC)**2) - 0.76*(1-SOC) + 4.10)
    else:
        raise ValueError('Battery Type not recognized')

#%% Numba SimpleRPM functions (one for precalculated Voc one for t)
@njit(fastmath = True)
def SimpleRPMeqs_Voc(RPM, *args):
    '''
    rho, Voc precalculated from h and SOC (or given directly)
    TODO: full docstring
    '''
    Voc, Uinf, dT, rho, eta_c, eta_g, GR, rpm_list, coef_numba_prop_data, d, ns, np_batt, CB, Rb, BattType, KV, Rm, I0, nmot = args
    J = Uinf/((RPM/60)*d)
    CP = CPNumba(RPM, J, rpm_list, coef_numba_prop_data)
    Qm = (rho*((RPM/60)**2)*(d**5)*CP)/(2*np.pi*GR*eta_g)
    Im = Qm*KV*(np.pi/30) + I0
    Ib = (Im*nmot)/eta_c
    Vb = ns*Voc - Ib*Rb
    Vm = dT*Vb
    RPMcalc = KV*(Vm - Im*Rm)/GR
    return(RPMcalc, J, CP, Qm, Im, Ib, Vb, Vm)

@njit(fastmath = True)
def SimpleRPMeqs_t(RPM, *args):
    '''
    rho precalculated from h and t given to calculate SOC, Voc
    TODO: full docstring
    '''
    t, Uinf, dT, rho, eta_c, eta_g, GR, rpm_list, coef_numba_prop_data, d, ns, np_batt, CB, Rb, BattType, KV, Rm, I0, nmot = args
    J = Uinf/((RPM/60)*d)
    CP = CPNumba(RPM, J, rpm_list, coef_numba_prop_data)
    Qm = (rho*((RPM/60)**2)*(d**5)*CP)/(2*np.pi*GR*eta_g)
    Im = Qm*KV*(np.pi/30) + I0
    Ib = (Im*nmot)/eta_c
    SOC = 1.0 - (Ib*t)/(3.6*CB*np_batt)
    Vb = ns*VocFunc(SOC, BattType) - Ib*Rb
    Vm = dT*Vb
    RPMcalc = KV*(Vm - Im*Rm)/GR
    return(RPMcalc, J, CP, Qm, Im, Ib, Vb, Vm)

# I need some method of setting whether we're inputting h or rho/t or SOC or Voc for the same function
@njit(fastmath = True)
def SimplifiedRPM_Voc(Uinf, dT, rho, Voc, *args):
    '''
    rho given directly or precalculated from h
    Voc given directly or precalculated from SOC
    TODO: full docstring
    
    Output indexes:
    0  1   2       3        4      5      6      7       8    
    T, Q, RPM, eta_drive, eta_p, eta_g, eta_m, eta_c, eta_b, 
    
     9      10     11    12     13    14   15  16  17  18  19  20   21   22
    Pout, Pin_m, Pin_c, Pw_m   Pw_c  Pw_b  Im, Ic, Ib, Vm, Vc, Vb, Voc, SOC
    '''
    ##### Assume ESC efficiency = 0.93 (93%) ##### 
    # note: max ESC efficiency occurs at dT = 1.0
    eta_c = 0.93
    
    # in the future could optimize parameters passed for memory, for now, it aids readability to pass everything
    GR, rpm_list, coef_numba_prop_data, d, ns, np, CB, Rb, BattType, KV, Rm, I0, nmot = args
    
    # Gearing efficiency
    if GR != 1.0:
        eta_g = 0.94 # 94% gear efficiency assumed
    else:
        eta_g = 1.0

    def residualfunc(RPM, *args):
        RPMcalc = SimpleRPMeqs_Voc(RPM, *args)[0]
        res = RPMcalc - RPM
        return(res)
    
    # Vbattinit = throttle*3.6*ns
    high = rpm_list[-1]
    low = rpm_list[0]
    RPM = bisection(low, high, residualfunc, Voc, Uinf, dT, rho, eta_c, eta_g, *args)
    
    _, J, CP, Q, Im, Ib, Vb, Vm = SimpleRPMeqs_Voc(RPM, Voc, Uinf, dT, rho, eta_c, eta_g, *args)
    if CP == 0.0:
        return([0.0]*23)
    CT = CTNumba(RPM, J, rpm_list, coef_numba_prop_data) 
    Ic = Ib/nmot 
    Vc = Vb 
    
    Pout =  rho*((RPM/60)**3)*(d**5)*CP     # mechanical power out of ONE motor
    Pin_m = Vm*Im                           # electric power into one motor
    Pin_c = Vc*Ic                           # electric power into one controller
    
    # Waste power calculations (via power loss); useful for heat estimation
    Pw_m = Pin_m - Pout
    Pw_c = Pin_c - Pin_m
    Pw_b = Rb*(Ib**2)
    
    eta_p = (CT*J)/CP
    eta_m = Pout/Pin_m 
    eta_c = Pin_m/Pin_c 
    eta_b = 1.0 - (Ib**2)*Rb/(nmot*Pin_c + ((Ib**2)*Rb))
    eta_drive = (eta_p*eta_g*eta_m*eta_c)*eta_b
    
    T = nmot*rho*((RPM/60)**2)*(d**4)*CT
    Q *= nmot
    
    # Calculate SOC via bisection between 0, 1
    # @njit(fastmath = True)
    def VocResidual(SOC, *args):
        Voc, BattType = args
        return(VocFunc(SOC, BattType) - Voc)
    SOC = bisection(0, 1, VocResidual, Voc, BattType)
    
    return([T, Q, RPM, 
            eta_drive, eta_p, eta_g, eta_m, eta_c, eta_b, 
            Pout, Pin_m, Pin_c, Pw_m, Pw_c, Pw_b,
            Im, Ic, Ib, Vm, Vc, Vb, Voc, SOC])


# I need some method of setting whether we're inputting h or rho/t or SOC or Voc for the same function
@njit(fastmath = True)
def SimplifiedRPM_t(Uinf, dT, rho, t, *args):
    '''
    rho given directly or precalculated from h
    t given directly
    
    Output indexes:
    0  1   2       3        4      5      6      7       8    
    T, Q, RPM, eta_drive, eta_p, eta_g, eta_m, eta_c, eta_b, 
    
     9      10     11    12     13    14   15  16  17  18  19  20   21   22
    Pout, Pin_m, Pin_c, Pw_m   Pw_c  Pw_b  Im, Ic, Ib, Vm, Vc, Vb, Voc, SOC
    '''
    ##### Assume ESC efficiency = 0.93 (93%) ##### 
    # note: max ESC efficiency occurs at dT = 1.0
    eta_c = 0.93
    
    # in the future could optimize parameters passed for memory, for now, it aids readability to pass everything
    GR, rpm_list, coef_numba_prop_data, d, ns, np, CB, Rb, BattType, KV, Rm, I0, nmot = args
    
    # Gearing efficiency
    if GR != 1.0:
        eta_g = 0.94 # 94% gear efficiency assumed
    else:
        eta_g = 1.0

    def residualfunc(RPM, *args):
        RPMcalc = SimpleRPMeqs_t(RPM, *args)[0]
        res = RPMcalc - RPM
        return(res)
    
    # Vbattinit = throttle*3.6*ns
    high = rpm_list[-1]
    low = rpm_list[0]
    RPM = bisection(low, high, residualfunc, t, Uinf, dT, rho, eta_c, eta_g, *args)
    
    _, J, CP, Q, Im, Ib, Vb, Vm = SimpleRPMeqs_t(RPM, t, Uinf, dT, rho, eta_c, eta_g, *args)
    if CP == 0.0:
        return([0.0]*23)
    CT = CTNumba(RPM, J, rpm_list, coef_numba_prop_data) 
    Ic = Ib/nmot 
    Vc = Vb 
    
    Pout =  rho*((RPM/60)**3)*(d**5)*CP     # mechanical power out of ONE motor
    Pin_m = Vm*Im                           # electric power into one motor
    Pin_c = Vc*Ic                           # electric power into one controller
    
    # Waste power calculations (via power loss); useful for heat estimation
    Pw_m = Pin_m - Pout
    Pw_c = Pin_c - Pin_m
    Pw_b = Rb*(Ib**2)
    
    eta_p = (CT*J)/CP
    eta_m = Pout/Pin_m 
    eta_c = Pin_m/Pin_c 
    eta_b = 1.0 - (Ib**2)*Rb/(nmot*Pin_c + ((Ib**2)*Rb))
    eta_drive = (eta_p*eta_g*eta_m*eta_c)*eta_b
    
    SOC = 1.0 - (Ib*t)/(3.6*CB*np)
    Voc = VocFunc(SOC, BattType)
    
    T = nmot*rho*((RPM/60)**2)*(d**4)*CT
    Q *= nmot
    return([T, Q, RPM, 
            eta_drive, eta_p, eta_g, eta_m, eta_c, eta_b, 
            Pout, Pin_m, Pin_c, Pw_m, Pw_c, Pw_b,
            Im, Ic, Ib, Vm, Vc, Vb, Voc, SOC])


#%% Non-Numba SimpleRPM functions (one for precalculated Voc one for t)
def VocFuncBase(SOC, BattType):
    '''
    Determines the battery Voltage (Voc) as a function of State of Charge (SOC) given from 0-1 for a specified battery chemistry
    
    LiPo
    Main equation from Chen 2006
    https://rincon-mora.gatech.edu/publicat/jrnls/tec05_batt_mdl.pdf
    
    Alternative from Jeong 2020
    Voc = 1.7*(SOC**3) - 2.1*(SOC**2) + 1.2*SOC + 3.4
    https://www.researchgate.net/publication/347270768_Improvement_of_Electric_Propulsion_System_Model_for_Performance_Analysis_of_Large-Size_Multicopter_UAVs
    
    NOTE:
    Jeong 2020 also presents resistance as a function of cell energy:
    Rbatt.cell = 21.0*(ebatt.cell)**-0.8056
    
    Liion
    https://www.researchgate.net/publication/346515863_Comparison_of_Lithium-Ion_Battery_Pack_Models_Based_on_Test_Data_from_Idaho_and_Argonne_National_Laboratories
    Using the i3 battery pack as an intermediate value
    
    TODO: implment battery voltage equation adjustments based on predicted health (measured via Vsoc at full charge!)
    '''
    if BattType == 'LiPo':
        return(3.685 - 1.031 * np.exp(-35 * SOC) + 0.2156 * SOC - 0.1178 * SOC**2 + 0.3201 * SOC**3)
    elif BattType == 'Liion':
        return(-4.48*((1-SOC)**5) + 9.09*((1-SOC)**4) - 7.08*((1-SOC)**3) + 2.32*((1-SOC)**2) - 0.76*(1-SOC) + 4.10)
    else:
        raise ValueError('Battery Type not recognized')

def SimpleRPMeqsBase_Voc(RPM, *args):
    '''
    rho, Voc precalculated from h and SOC (or given directly)
    TODO: full docstring
    '''
    Voc, Uinf, dT, rho, eta_c, eta_g, GR, rpm_list, coef_numba_prop_data, d, ns, np_batt, CB, Rb, BattType, KV, Rm, I0, nmot = args
    J = Uinf/((RPM/60)*d)
    CP = CPBase(RPM, J, rpm_list, coef_numba_prop_data)
    Qm = (rho*((RPM/60)**2)*(d**5)*CP)/(2*np.pi*GR*eta_g)
    Im = Qm*KV*(np.pi/30) + I0
    Ib = (Im*nmot)/eta_c
    Vb = ns*Voc - Ib*Rb
    Vm = dT*Vb
    RPMcalc = KV*(Vm - Im*Rm)/GR
    return(RPMcalc, J, CP, Qm, Im, Ib, Vb, Vm)

def SimpleRPMeqsBase_t(RPM, *args):
    '''
    rho precalculated from h and t given to calculate SOC, Voc
    
    Output indexes:
    0  1   2       3        4      5      6      7       8    
    T, Q, RPM, eta_drive, eta_p, eta_g, eta_m, eta_c, eta_b, 
    
     9      10     11    12     13    14   15  16  17  18  19  20   21   22
    Pout, Pin_m, Pin_c, Pw_m   Pw_c  Pw_b  Im, Ic, Ib, Vm, Vc, Vb, Voc, SOC
    '''
    t, Uinf, dT, rho, eta_c, eta_g, GR, rpm_list, coef_numba_prop_data, d, ns, np_batt, CB, Rb, BattType, KV, Rm, I0, nmot = args
    J = Uinf/((RPM/60)*d)
    CP = CPBase(RPM, J, rpm_list, coef_numba_prop_data)
    Qm = (rho*((RPM/60)**2)*(d**5)*CP)/(2*np.pi*GR*eta_g)
    Im = Qm*KV*(np.pi/30) + I0
    Ib = (Im*nmot)/eta_c
    SOC = 1.0 - (Ib*t)/(3.6*CB*np_batt)
    Vb = ns*VocFuncBase(SOC, BattType) - Ib*Rb
    Vm = dT*Vb
    RPMcalc = KV*(Vm - Im*Rm)/GR
    return(RPMcalc, J, CP, Qm, Im, Ib, Vb, Vm)

# I need some method of setting whether we're inputting h or rho/t or SOC or Voc for the same function
def SimplifiedRPMBase_Voc(Uinf, dT, rho, Voc, *args):
    '''
    rho given directly or precalculated from h
    Voc given directly or precalculated from SOC
    
    Output indexes:
    0  1   2       3        4      5      6      7       8    
    T, Q, RPM, eta_drive, eta_p, eta_g, eta_m, eta_c, eta_b, 
    
     9      10     11    12     13    14   15  16  17  18  19  20   21   22
    Pout, Pin_m, Pin_c, Pw_m   Pw_c  Pw_b  Im, Ic, Ib, Vm, Vc, Vb, Voc, SOC
    '''
    ##### Assume ESC efficiency = 0.93 (93%) ##### 
    # note: max ESC efficiency occurs at dT = 1.0
    eta_c = 0.93
    
    # in the future could optimize parameters passed for memory, for now, it aids readability to pass everything
    GR, rpm_list, coef_numba_prop_data, d, ns, np_batt, CB, Rb, BattType, KV, Rm, I0, nmot = args
    
    # Gearing efficiency
    if GR != 1.0:
        eta_g = 0.94 # 94% gear efficiency assumed
    else:
        eta_g = 1.0

    def residualfunc(RPM, *args):
        RPMcalc = SimpleRPMeqsBase_Voc(RPM, *args)[0]
        res = RPMcalc - RPM
        return(res)
    
    high = rpm_list[-1]
    low = rpm_list[0]
    RPM = bisectionBase(low, high, residualfunc, Voc, Uinf, dT, rho, eta_c, eta_g, *args)
    
    _, J, CP, Q, Im, Ib, Vb, Vm = SimpleRPMeqsBase_Voc(RPM, Voc, Uinf, dT, rho, eta_c, eta_g, *args)
    if CP == 0.0:
        return([0.0]*23)
    CT = CTBase(RPM, J, rpm_list, coef_numba_prop_data) 
    Ic = Ib/nmot 
    Vc = Vb 
    
    Pout =  rho*((RPM/60)**3)*(d**5)*CP     # mechanical power out of ONE motor
    Pin_m = Vm*Im                           # electric power into one motor
    Pin_c = Vc*Ic                           # electric power into one controller
    
    # Waste power calculations (via power loss); useful for heat estimation
    Pw_m = Pin_m - Pout
    Pw_c = Pin_c - Pin_m
    Pw_b = Rb*(Ib**2)
    
    eta_p = (CT*J)/CP
    eta_m = Pout/Pin_m 
    eta_c = Pin_m/Pin_c 
    eta_b = 1.0 - ((Ib**2)*Rb)/(nmot*Pin_c + ((Ib**2)*Rb))
    eta_drive = eta_p*eta_g*eta_m*eta_c*eta_b # took out divided by nmot, REVIEW later
    
    T = nmot*rho*((RPM/60)**2)*(d**4)*CT
    Q *= nmot
    
    # Calculate SOC via bisection between 0, 1
    def VocResidual(SOC):
        return(VocFuncBase(SOC, BattType) - Voc)
    SOC = bisectionBase(0, 1, VocResidual)
    if SOC < 0:
        raise ValueError('SOC < 0')
    
    return([T, Q, RPM, 
            eta_drive, eta_p, eta_g, eta_m, eta_c, eta_b, 
            Pout, Pin_m, Pin_c, Pw_m, Pw_c, Pw_b,
            Im, Ic, Ib, Vm, Vc, Vb, Voc, SOC])


# I need some method of setting whether we're inputting h or rho/t or SOC or Voc for the same function
def SimplifiedRPMBase_t(Uinf, dT, rho, t, *args):
    '''
    rho given directly or precalculated from h
    t given directly
    
    Output indexes:
    0  1   2       3        4      5      6      7       8    
    T, Q, RPM, eta_drive, eta_p, eta_g, eta_m, eta_c, eta_b, 
    
     9      10     11    12     13    14   15  16  17  18  19  20   21   22
    Pout, Pin_m, Pin_c, Pw_m   Pw_c  Pw_b  Im, Ic, Ib, Vm, Vc, Vb, Voc, SOC
    '''
    ##### Assume ESC efficiency = 0.93 (93%) ##### 
    # note: max ESC efficiency occurs at dT = 1.0
    eta_c = 0.93
    
    # in the future could optimize parameters passed for memory, for now, it aids readability to pass everything
    GR, rpm_list, coef_numba_prop_data, d, ns, np_batt, CB, Rb, BattType, KV, Rm, I0, nmot = args
    
    # Gearing efficiency
    if GR != 1.0:
        eta_g = 0.94 # 94% gear efficiency assumed
    else:
        eta_g = 1.0

    def residualfunc(RPM, *args):
        RPMcalc = SimpleRPMeqsBase_t(RPM, *args)[0]
        res = RPMcalc - RPM
        return(res)
    
    # Vbattinit = throttle*3.6*ns
    high = rpm_list[-1]
    low = rpm_list[0]
    RPM = bisectionBase(low, high, residualfunc, t, Uinf, dT, rho, eta_c, eta_g, *args)
    
    _, J, CP, Q, Im, Ib, Vb, Vm = SimpleRPMeqsBase_t(RPM, t, Uinf, dT, rho, eta_c, eta_g, *args)
    if CP == 0.0:
        return([0.0]*23)
        
    CT = CTBase(RPM, J, rpm_list, coef_numba_prop_data)
    Ic = Ib/nmot 
    Vc = Vb 
    
    Pout =  rho*((RPM/60)**3)*(d**5)*CP     # mechanical power out of ONE motor
    Pin_m = Vm*Im                           # electric power into one motor
    Pin_c = Vc*Ic                           # electric power into one controller
    
    # Waste power calculations (via power loss); useful for heat estimation
    Pw_m = Pin_m - Pout
    Pw_c = Pin_c - Pin_m
    Pw_b = Rb*(Ib**2)

    eta_p = (CT*J)/CP
    eta_m = Pout/Pin_m 
    eta_c = Pin_m/Pin_c 
    eta_b = 1.0 - ((Ib**2)*Rb)/(nmot*Pin_c + ((Ib**2)*Rb)) 
    # TODO: FIX THE BATTERY EFFICIENCY BY ACCOUNTING FOR RB VARYING WITH SOC (CRITICAL)
    # Check battery efficiency varies with runtime: 
    # https://ntrs.nasa.gov/api/citations/20205004497/downloads/Battery_Evaluation_EATS_07_15_20.pdf
    eta_drive = eta_p*eta_g*eta_m*eta_c*eta_b # took out divided by nmot, REVIEW later
    
    T = nmot*rho*((RPM/60)**2)*(d**4)*CT
    Q *= nmot
    
    SOC = 1.0 - (Ib*t)/(3.6*CB*np_batt)
    Voc = VocFuncBase(SOC, BattType)
    
    return([T, Q, RPM, 
            eta_drive, eta_p, eta_g, eta_m, eta_c, eta_b, 
            Pout, Pin_m, Pin_c, Pw_m, Pw_c, Pw_b,
            Im, Ic, Ib, Vm, Vc, Vb, Voc, SOC])


#%%################## PLOTTING FUNCTIONS ##################

################################# PointResult #################################
def exactly_one_defined(*args) -> bool:
    return sum(x is not None for x in args) == 1

def PointResultFunc(self, Uinf = None, dT = None, 
                    rho = None, h = None, 
                    SOC = None, Voc = None, t = None, 
                    verbose = True):
    ''' 
    PointResult for fixed Uinf, dT, t/SOC/Voc, h/rho
        
    Outputs
    --------------------------------
    array of:
    0  1   2       3        4      5      6      7       8    
    T, Q, RPM, eta_drive, eta_p, eta_g, eta_m, eta_c, eta_b, 
    
     9      10     11    12     13    14   15  16  17  18  19  20   21   22
    Pout, Pin_m, Pin_c, Pw_m   Pw_c  Pw_b  Im, Ic, Ib, Vm, Vc, Vb, Voc, SOC
    
    '''
    args = (self.GR, self.rpm_list, self.COEF_NUMBA_PROP_DATA, self.propdiam, 
            self.ns, self.np, self.CB, self.Rb, self.BattType, 
            self.KV, self.Rm, self.I0, self.nmot)
    if not exactly_one_defined(t, SOC, Voc):
        raise ValueError("Exactly one of t, SOC, or Voc must be provided")
        
    if not exactly_one_defined(rho, h):
        raise ValueError("Exactly one of h or rho must be provided")
    
    if h is not None:
        rho = atm().rho(h)
                
    if t is not None: 
        try:
            propQs = SimplifiedRPMBase_t(Uinf, dT, rho, t, *args)
        except:
            print('ERROR: Infeasible input combination, try reducing runtime')
            return(np.zeros(23))
        if propQs[22] < 0.0:
            print('ERROR: SOC < 0% for that runtime')
            return(np.zeros(23))
    else:
        # Voc or SOC input
        if SOC is not None:
            Voc = VocFuncBase(SOC, self.BattType)
        try:
            propQs = SimplifiedRPMBase_Voc(Uinf, dT, rho, Voc, *args)
        except:
            print('ERROR: Infeasible input combination, try reducing runtime')
            return(np.zeros(23))
        
    if verbose:
        if t is not None:
            print(f'\nAt Uinf = {Uinf/ftm:.2f} ft/s, Throttle = {dT*100:.0f}%, Density = {rho:.3f} kg/m3, Runtime = {t:.1f} s')
        elif SOC is not None and SOC > 0.0:
            print(f'\nAt Uinf = {Uinf/ftm:.2f} ft/s, Throttle = {dT*100:.0f}%, Density = {rho:.3f} kg/m3, SOC = {SOC*100:.0f}%')
        elif Voc is not None:
            print(f'\nAt Uinf = {Uinf/ftm:.2f} ft/s, Throttle = {dT*100:.0f}%, Density = {rho:.3f} kg/m3, Voc = {Voc:.2f} V')

        for i, name in enumerate(propQnames):
            if name == 'Total Thrust (lbf)':
                print(f'{name:30} = {propQs[i]/lbfN:.3f}')
            elif 'Efficiency' in name:
                print(f'{name:30} = {propQs[i]*100:.2f}%')
            elif 'State of Charge' in name:
                print(f'{name:30} = {propQs[i]*100:.2f}%')
            else:
                print(f'{name:30} = {propQs[i]:.3f}')
    return(np.array(propQs))

#%%################################ LinePlot #################################
# @njit(fastmath = True)
# def process_LinePlot_Vinf_SOC(Vs, SOC_Voc, dT, propQindx, *args):
#     ''' To allow for numbification without needing to recompile every iteration '''
#     outputs = []
#     thrusts = []
#     for Vinf in Vs:
#         propQ = SimplifiedRPM(Vinf, dT, SOC_Voc, *args)
#         outputs.append(propQ[propQindx])
#         thrusts.append(propQ[0])
#     return(outputs, thrusts)

# @njit(fastmath = True)
# def process_LinePlot_Vinf_t(Vs, t, dT, propQindx, *args):
#     ''' Identical to above function but uses the constant current t approximation '''
#     outputs = []
#     thrusts = []
#     for Vinf in Vs:
#         propQ = SimplifiedRPM_t(Vinf, dT, t, *args)
#         outputs.append(propQ[propQindx])
#         thrusts.append(propQ[0])
#     return(outputs, thrusts)

def LinePlotFunc(self, propQ = 'T',
                 Uinf = None, dT = None, 
                 rho = None, h = None, 
                 SOC = None, Voc = None, t = None, 
                 verbose = True, plot = False):
    '''
    Input
    ----------------------------------------------------------------------------------------------------------
        a propulsion quantity (propQ) of interest (options given by the output array)
        
        constant values for three of: Uinf, dT, rho/h, SOC/Voc/t
        a range of the final value 
        
    IMPORTANT: 
        bounds on ranges: dT in (0, 1), rho >= 0, h >= 0, SOC in (0, 1), Voc in (2.0, 4.2), t >= 0

    Output
    ----------------------------------------------------------------------------------------------------------
        2D np array with columns corresponding to 
        
        0  1   2       3        4      5      6      7       8    
        T, Q, RPM, eta_drive, eta_p, eta_g, eta_m, eta_c, eta_b, 
        
         9      10     11    12    13    14   15  16  17  18  19  20   21   22
        Pout, Pin_m, Pin_c, Pw_m  Pw_c  Pw_b  Im, Ic, Ib, Vm, Vc, Vb, Voc, SOC
        
        and rows corresponding to a range of the nonconstant value
    '''
    
    args = (self.GR, self.rpm_list, self.COEF_NUMBA_PROP_DATA, self.propdiam, 
            self.ns, self.np, self.CB, self.Rb, self.BattType, 
            self.KV, self.Rm, self.I0, self.nmot)
    
    if not exactly_one_defined(t, SOC, Voc):
        raise ValueError("Exactly one of t, SOC, or Voc must be provided")
        
    if not exactly_one_defined(rho, h):
        raise ValueError("Exactly one of h or rho must be provided")
    
    # Find the np array and the constants to use
    full_inputs = [Uinf, dT, rho, h, SOC, Voc, t]
    full_input_names = ['Velocity (m/s)', 'Throttle (0-1)', 
                        'Density (kg/m\u00B3)', 'Altitude (m)', 
                        'State of Charge (0-1)', 'Cell Voltage (V)', 'Runtime (s)']
    
    arrs = 0
    idxarr = 0
    for i, specinput in enumerate(full_inputs):
        if isinstance(specinput, np.ndarray):
            arrs += 1
            idxarr = i
            inputarr = specinput
    if arrs != 1:
        raise KeyError("Exactly one array of inputs must be provided")
    
    input_name = full_input_names[idxarr]
    
    # convert h to density
    if h is not None:
        rho = atm().rho(h)
        
    # convert SOC to Voc
    if SOC is not None:
        Voc = VocFuncBase(SOC, self.BattType)
    
    # collapsing indexes
    if idxarr == 3:
        idxarr = 2 
    elif idxarr == 4 or idxarr == 5 or idxarr == 6:
        idxarr = 3 

    if t is None:
        clean_inputs = [Uinf, dT, rho, Voc] # if Voc is arr, SOC is arr and vice versa
        PropQs = np.zeros((clean_inputs[idxarr].size, 23))
        for i, value in enumerate(clean_inputs[idxarr]):
            inner_input = copy.deepcopy(clean_inputs)
            inner_input[idxarr] = value
            PropQs[i, :] = SimplifiedRPM_Voc(inner_input[0], inner_input[1], inner_input[2], inner_input[3], *args)
            # TODO add speedup based on when the data starts and ends 
            # (assume it has no discontinuities in between!)

            # if PropQs[i, 0] <= 0: # indicates that the solution is infeasible
            #     endidx = i 
            #     break
        endidx = clean_inputs[idxarr].size
    else:
        clean_inputs = [Uinf, dT, rho, t]
        PropQs = np.zeros((clean_inputs[idxarr].size, 23))
        inner_input = copy.deepcopy(clean_inputs)
        for i, value in enumerate(clean_inputs[idxarr]):
            inner_input[idxarr] = value
            PropQs[i, :] = SimplifiedRPM_t(inner_input[0], inner_input[1], inner_input[2], inner_input[3], *args)
            # todo, add speedup based on when the data starts and ends 
            # (assume it has no discontinuities in between!)
            # if PropQs[i, 0] <= 0: # indicates that the solution is infeasible
            #     endidx = i 
            #     break
            if PropQs[i, -1] <= 1-self.ds: # minimum SOC = 10% for batteries
                endidx = i
                PropQs[i, :] = np.array([0.0]*23)
        endidx = clean_inputs[idxarr].size
    inputarr = inputarr[:endidx]
    PropQs = PropQs[:endidx, :]
    
    if plot:
        # primary goal: plot the inputarray on the x-axis versus a selected propQ on the yaxis
        # secondary goal: if some aircraft characteristics (Cd, Sw) have been provided, plot the T = D over the selected propQ
        # anything else? Do I want to have an option of plotting a contour where propQ is yaxis, and another var (such as dT corresponds to different contour lines?)
        # I say no for now

        # IMPORTANT: figure out what to do with the values that don't converge
        # do I plot them directly and let people alter their inputs?
        # would prefer to get the code to determine the bounds automatically and raise errors
        #   if the range is extremely limited
        
        # TODO: add in Ilimit, Plimit (from motor/battery/ESC database)
        # TODO: add in T = D
        # TODO: limit plot (or data gathering range automatically 
        # so ppl don't have to input values 
        # they just specify which variable they want to be the array (simple)
        
        propqidx = propQshort.index(propQ)

        fig, ax = plt.subplots(figsize = (6, 4))
        ax.plot(inputarr, PropQs[:, propqidx], color = 'k')
        plt.xlabel(input_name)
        plt.ylabel(propQnames[propqidx])
        
        # get short names of inputs along with title string
        input_names = ['Uinf', 'dT', 'rho', 'h', 'SOC', 'Voc', 't']
        parts = []
        for name, val in zip(input_names, full_inputs):
            if val is None:
                continue
            if isinstance(val, np.ndarray):
                continue  # skip arrays
            parts.append(f"{name} = {val}")
        title_str = ", ".join(parts)
        plt.title(f'{input_name} sweep; {title_str}' + f'\n{self.nmot} {self.motor_name} motor, {self.prop_name} propeller, {self.batt_name} battery')
        plt.show()

    
    return(PropQs, inputarr)


#%% ContourPlot
# @njit(fastmath = True)
# def process_2D_data_Voc():
    
#     return()

# @njit(fastmath = True)
# def process_2D_data_t():
    
#     return()

# can plot any two of Uinf, dT, h/rho, t/Voc/SOC
def ContourPlotFunc(self, propQ = 'T',
                    xaxis = None,
                    yaxis = None,
                    Uinf = None, dT = None, 
                    rho = None, h = None, 
                    SOC = None, Voc = None, t = None, 
                    verbose = True, plot = False,
                    colormap = 'viridis', 
                    grade = 30):
    '''
    Input
    ----------------------------------------------------------------------------------------------------------
        a propulsion quantity (propQ) of interest (options given by the output array)
        xaxis, yaxis for plot specified from the names:
            ['SOC', 'Voc', 't', 'Uinf', 'dT', 'rho', 'h']
        
        constant values for two of: 
            SOC/Voc/t, Uinf, dT, rho/h
        np.array for the other two values!
        
    IMPORTANT: 
        bounds on ranges: dT in (0, 1), rho >= 0, h >= 0, SOC in (0, 1), Voc in (2.0, 4.2), t >= 0
        input arrays MUST have the same size (square input)

    Output
    ----------------------------------------------------------------------------------------------------------
        3D np array (0, 0, :) corresponding to 
        
        0  1   2       3        4      5      6      7       8    
        T, Q, RPM, eta_drive, eta_p, eta_g, eta_m, eta_c, eta_b, 
        
         9      10     11    12    13    14   15  16  17  18  19  20   21   22
        Pout, Pin_m, Pin_c, Pw_m  Pw_c  Pw_b  Im, Ic, Ib, Vm, Vc, Vb, Voc, SOC
        
        (0, :, 0) corresponding to the x axis input
        (:, 0, 0) corresponding to the y axis input
        
        
    Little note:
        Might be confusing why input_names is in a different order compared to LinePlot. That's to provide
        better xaxis, yaxis selection by default (time on horizontal axis, altitude on vertical, etc)
    '''
    
    args = (self.GR, self.rpm_list, self.COEF_NUMBA_PROP_DATA, self.propdiam, 
            self.ns, self.np, self.CB, self.Rb, self.BattType, 
            self.KV, self.Rm, self.I0, self.nmot)
    
    if not exactly_one_defined(t, SOC, Voc):
        raise ValueError("Exactly one of t, SOC, or Voc must be provided")
        
    if not exactly_one_defined(rho, h):
        raise ValueError("Exactly one of h or rho must be provided")
    
    full_inputs = [SOC, Voc, t, Uinf, dT, rho, h]
    input_names = ['SOC', 'Voc', 't', 'Uinf', 'dT', 'rho', 'h']

    # dictionary formulation
    provided_inputs = {name: val for name, val in zip(input_names, full_inputs) if val is not None}
    arrays = {}
    constants = {}
    
    for name, val in provided_inputs.items():
        if isinstance(val, (np.ndarray, list, tuple)):
            arrays[name] = np.asarray(val)
        else:
            constants[name] = val

    if len(arrays) != 2:
        raise ValueError(f"Expected exactly 2 array inputs, but found {len(arrays)}: {list(arrays.keys())}")

    if xaxis != None:
        if xaxis not in arrays:
            raise ValueError(f'The specified xaxis ({xaxis}) must be given as an array (i.e. Uinf = np.linspace(0, 30, 50))')
    else:
        xaxis = list(arrays)[0]
    if yaxis != None:
        if yaxis not in arrays:
            raise ValueError(f'The specified yaxis ({yaxis}) must be given as an array (i.e. Uinf = np.linspace(0, 30, 50))')
    else:
        yaxis = list(arrays)[1]
    
    if xaxis not in arrays or yaxis not in arrays:
        raise ValueError(f"The specified xaxis ('{xaxis}') and yaxis ('{yaxis}') must match the provided array inputs.")
    
    x_array = arrays[xaxis]
    y_array = arrays[yaxis]
    
    if x_array.shape != y_array.shape:
        raise ValueError(f"Input arrays must have the same size. Got {x_array.shape} and {y_array.shape}.")

    grid_size = len(x_array)
    output_array = np.zeros((grid_size, grid_size, 23))

    for i, y_val in enumerate(tqdm(y_array)):
        for j, x_val in enumerate(x_array):
            
            current_state = constants.copy()
            current_state[xaxis] = x_val
            current_state[yaxis] = y_val
            
            curr_Uinf = current_state['Uinf']
            curr_dT = current_state['dT']
            
            if 'rho' in current_state:
                curr_rho = current_state['rho']
            else:
                curr_rho = atm().rho(current_state['h'])
                
            # Select between Voc, SOC, t
            if 'Voc' in current_state:
                curr_Voc = current_state['Voc']
                output_vector = SimplifiedRPM_Voc(curr_Uinf, curr_dT, curr_rho, curr_Voc, *args)
            elif 'SOC' in current_state:
                curr_Voc =  VocFuncBase(current_state['SOC'], self.BattType) 
                output_vector = SimplifiedRPM_Voc(curr_Uinf, curr_dT, curr_rho, curr_Voc, *args)
            else:                
                #### t input
                curr_t = current_state['t']
                output_vector = SimplifiedRPM_t(curr_Uinf, curr_dT, curr_rho, curr_t, *args)
            
            output_array[i, j, :] = output_vector
        
    if plot:
        # TODO: add in Ilimit, Plimit (from motor/battery/ESC database)
        # TODO: add in T = D
        # TODO: limit plot (or data gathering range automatically 
        # so ppl don't have to input values 
        # they just specify which variable they want to be the array (simple)

        fig, ax = plt.subplots(figsize = (6, 4))
        
        propqidx = propQshort.index(propQ)
        propQ_spec = output_array[:, :, propqidx]
        
        X, Y = np.meshgrid(x_array, y_array)
        propQ_spec_alt = propQ_spec[propQ_spec > 0]    # very good for removing all the violation keys, and finding the max, but flattens the array
        lower = propQ_spec_alt.min()                    # finds the minimum Score discounting violation trips
        upper = propQ_spec.max()
        img = ax.contourf(X, Y, propQ_spec, cmap = colormap, levels = np.linspace(lower, upper, grade))
        fig.colorbar(img, ticks = np.linspace(lower, upper, 11), pad = 0.025, shrink = 1.0, spacing = 'uniform', label=f'{propQnames[propqidx]}')
    
        # to get correct names: xaxis, yaxis --> input names idx --> full_input_names idx --> full_input_names value
        full_input_names = ['State of Charge (0-1)', 'Cell Voltage (V)', 'Runtime (s)', 
                            'Velocity (m/s)', 'Throttle (0-1)', 
                            'Density (kg/m\u00B3)', 'Altitude (m)']
        xname_idx = 0
        yname_idx = 0
        for i, name in enumerate(input_names):
            if name == xaxis:
                xname_idx = i
            elif name == yaxis:
                yname_idx = i
        
        plt.xlabel(full_input_names[xname_idx])
        plt.ylabel(full_input_names[yname_idx])
        
        # get short names of inputs along with title string
        parts = []
        for name, val in zip(input_names, full_inputs):
            if val is None:
                continue
            if isinstance(val, np.ndarray):
                continue  # skip arrays
            parts.append(f"{name} = {val}")
        title_str = ", ".join(parts)
        plt.title(f'{xaxis}, {yaxis} sweeps; {title_str}' + f'\n{self.nmot} {self.motor_name} motor, {self.prop_name} propeller, {self.batt_name} battery')
        plt.show()
        
    
    # output array = (n, n, 23) where (:, 0, 0) corresponds to y and (0, :, :) corresponds to x
    # return xarr, yarr, output
    return(x_array, y_array, output_array)


# -*- coding: utf-8 -*-
"""
File dedicated to performance functions

@author: Sammy Nassau
"""

import numpy as np
import scipy
from scipy.integrate import solve_ivp
from scipy.integrate import trapezoid, cumulative_trapezoid, quad
import scipy.integrate as sp 
import matplotlib.pyplot as plt
from tqdm import tqdm
from numba import jit, njit

# import uavdex.propulsions as propulsions

lbfN = 4.44822
ftm = 0.3048

#%% Takeoff simulation using initial value problem methods
def SimulateTakeoff(self, aoa_rotation = 10, texpect = 200, results = True, plot = False):
    '''
    ASSUME ROTATION AT 4/5 OF VLOF
    
    aoa rotation is really finicky, ideally figure out what it would be irl and implement
    Assume 10 deg otherwise
    
    h0 = distance of fuselage centerline from ground (meters)
    b = wingspan (m)
    t expect is the expected takeoff time in s
    
    Currently no accounting for takeoff rotation!!
    
    Using: https://digitalcommons.usu.edu/cgi/viewcontent.cgi?article=1080&context=mae_facpub
    For more accurate ground effect predictions for wings with LINEAR TAPER
    
    '''
    print('\nSimulating Ground Roll...')
    # to account for ground effect (phillips + hunsaker correction) at SMALL aoa
    delta_D = 1 - 0.157*(self.taper**0.775 - 0.373)*(self.AR**0.417 - 1.27)
    CDige_oge = 1 - delta_D*(np.e**(-4.74*(self.h0/self.b)**0.814)) - ((self.h0/self.b)**2)*(np.e**(-3.88*(self.h0/self.b)**0.758))
    
    # at SMALL aerodynamic angles of attack
    delta_L = 1 - 2.25*(self.taper**0.00273 - 0.997)*(self.AR**0.717 + 13.6)
    CLige_oge = 1 + delta_L*(288*((self.h0/self.b)**0.787)*(np.e**(-9.14*((self.h0/self.b)**0.327))))/(self.AR**0.882)
    
    # Pre-rotation!
    M = self.MGTOW/9.81 # mass in kg
    
    def T_Dtakeoff(t, V):
        T, _, _, _ = propulsions.ModelCalcs(self, V, t) 
        D = 0.5*self.rho*(V**2)*self.Sw*self.CDtoPreR*CDige_oge
        L = 0.5*self.rho*(V**2)*self.Sw*self.CLtoPreR*CLige_oge
        F = self.mufric*(self.MGTOW-L)
        return((T - D - F)/M)
    
    V0 = [0]
    tspan= [0, texpect]
    
    sol = solve_ivp(T_Dtakeoff, tspan, V0, method = 'RK45', dense_output = True, t_eval=np.linspace(0, texpect, 100000)) #might need to change this if time becomes an issue
    ts = sol.t
    Vs = sol.y[0]
    
    diffs = np.abs(Vs - (5/6)*self.Vlof) # assume rotation at 4/5 Vlof (BIG BIG APPROX)
    rotate = np.argsort(diffs)[:1][0]
    # trotate = ts[rotate]
    xrotate = trapezoid(Vs[:rotate], x=ts[:rotate])
    
    t1 = ts[:rotate]
    V1 = Vs[:rotate]
    d1 = cumulative_trapezoid(V1, x=t1)
    d1 = np.insert(d1, 0, 0.0)
    a1 = Vs[1:rotate]/ts[1:rotate]
    
    # ROTATION phase
    # CLnew = self.CLmax-0.2 #HUGEEEE assumption, this would be changing continuously
    CLnew = self.CLtoPostR # at some deg of rotation with high lift devices
    
    # updated using the high aoa formulations from Phillips and Hunsaker
    Beta_D = 1 + 0.0361*(CLnew**1.21)/((self.AR**1.19)*((self.h0/self.b)**1.51))
    CDige_oge = (1 - delta_D*(np.e**(-4.74*(self.h0/self.b)**0.814)) - ((self.h0/self.b)**2)*(np.e**(-3.88*(self.h0/self.b)**0.758)))*Beta_D
    Beta_L = 1 + 0.269*(CLnew**1.45)
    CLige_oge = (1 + (delta_L*(288*(self.h0/self.b)**0.787)*(np.e**(-9.14*((self.h0/self.b)**0.327))))/(self.AR**0.882))/Beta_L
    
    # assume that the takeoff occurs at 10 deg aoa, and adjust the thrust vector accordingly
    aoa = aoa_rotation #deg
    def T_Dtakeoff(t, V):
        T, _, _, _ = propulsions.ModelCalcs(self, V, t)
        T *= np.cos(aoa*(np.pi/180))
        D = 0.5*self.rho*(V**2)*self.Sw*self.CDtoPostR*CDige_oge
        L = 0.5*self.rho*(V**2)*self.Sw*self.CLtoPostR*CLige_oge + T*np.sin(aoa*(np.pi/180))
        F = self.mufric*(self.MGTOW-L)
        return((T - D - F)/M)
    
    V0 = [V1[-1]]
    tspan= [t1[-1], texpect]
    
    sol = solve_ivp(T_Dtakeoff, tspan, V0, method = 'RK45', dense_output = True, t_eval=np.linspace(t1[-1], texpect, 100000)) #might need to change this if time becomes an issue
    ts = sol.t
    Vs = sol.y[0]
    
    diffs = np.abs(Vs - self.Vlof) # assume rotation at 2/3 Vlof (BIG BIG APPROX)
    liftoff = np.argsort(diffs)[:1][0]
    tlof = ts[liftoff]
    
    t2 = ts[:liftoff]
    V2 = Vs[:liftoff]
    d2 = cumulative_trapezoid(V2, x=t2) + d1[-1]
    d2 = np.insert(d2, 0, d1[-1])
    a2 = Vs[:liftoff]/ts[:liftoff]
    xlof = trapezoid(Vs[:liftoff], x=ts[:liftoff])
    
    t_tot = np.append(t1, t2)
    V_tot = np.append(V1, V2)
    d_tot = np.append(d1, d2)
    a_tot = np.append(a1, a2)
    xlof_tot = xrotate + xlof
    tlof_tot = tlof
    if plot:
        fig, ax1 = plt.subplots(figsize = (6, 4), dpi = 1000)
        
        p1 = ax1.plot(t_tot, V_tot/ftm, color = 'black')
        ax1.set_ylabel('Velocity (ft/s)')
        
        ax2 = ax1.twinx()
        p2 = ax2.plot(t_tot, d_tot/ftm, color = 'red')
        ax2.set_ylabel('Distance (ft)')
        
        ax3 = ax1.twinx()
        p3 = ax3.plot(t_tot[1:], a_tot/ftm, color = 'blue')
        ax3.set_ylabel(r'Acceleration (ft/s$^2$)')
        ax3.spines['right'].set_position(('outward', 60))

        plt.xlabel('Time (s)')
        
        ax1.plot([t2[0], t2[0]], ax1.get_ylim(), '--', color = 'orange', label = 'Rotation')
        
        ax1.yaxis.label.set_color(p1[0].get_color())
        ax2.yaxis.label.set_color(p2[0].get_color())
        ax3.yaxis.label.set_color(p3[0].get_color())

        ax1.grid()
        ax1.minorticks_on()
        ax2.minorticks_on()
        ax3.minorticks_on()
        ax1.legend(loc = 'upper center')
        plt.annotate('Rotation', (1, 1))
        ax1.set_xlabel('Time (s)')
        plt.title(f'Ground Roll, MGTOW = {self.MGTOW/lbfN:.1f} lbf\nVlof = {self.Vlof/ftm:.1f} fps, Vrotate = {V1[-1]/ftm:.1f} ft/s\nDistance = {xlof_tot/ftm:.1f} ft, Time = {tlof_tot:.2f} s')
        plt.show()
    
    if results:
        print('\nGround Roll Calculations:\n'
              f'{"Distance":10} {xlof_tot/ftm:.4f} ft\n'
              f'{"Time":10} {tlof_tot:.4f} s')
    return(tlof_tot, xlof_tot)

#%% The real functions start here!
@njit #DOESN'T WORK BC OF THE TRAPEZOID USAGE!!!, NOW IT DOES!!
def Takeoff(aoa, texpect, h0, taper, AR, b, MGTOW, rho, Sw, CDtoPreR, CLtoPreR, CDtoPostR, CLtoPostR, CLmax, 
            mass, mufric, Vlof, rpm_list, NUMBA_PROP_DATA, CB, ns, Rint, KV, Rm, nmot, I0, ds, takeoffdT, GR, m = 1000, plot = False, results = False):
    '''
    
    Several approximations in this code:
        - the phillips + hunsaker correction is valid for ground affect impact for very small UAVs
        - ***rotation occurs instantaneously at 4/5 of liftoff speed (itself 1.15*Vstall)
        - g = 9.81 m/s^2
        
    '''    
    h = h0
    # to account for ground effect (phillips + hunsaker correction) at SMALL aoa
    delta_D = 1 - 0.157*(taper**0.775 - 0.373)*(AR**0.417 - 1.27)
    CDige_oge = 1 - delta_D*(np.e**(-4.74*(h/b)**0.814)) - ((h/b)**2)*(np.e**(-3.88*(h/b)**0.758))
    
    # at SMALL aerodynamic angles of attack
    delta_L = 1 - 2.25*(taper**0.00273 - 0.997)*(AR**0.717 + 13.6)
    CLige_oge = 1 + delta_L*(288*((h/b)**0.787)*(np.e**(-9.14*((h/b)**0.327))))/(AR**0.882)
    
    
    # timestepping from mission start to rotate velocity assumption
    ts = np.linspace(0.0, texpect, m) # start the mission at 0.0 s
    dt = ts[2]-ts[1]
    
    x = np.zeros(m+1)
    V = np.zeros(m+1)
    SOC = np.zeros(m+1)
    V[0] = 0.0 #initial V (m/s)
    SOC[0] = 1.0 #initial state of charge

    a = np.zeros(m)
    T = np.zeros(m)
    P = np.zeros(m)
    Itot = np.zeros(m)
    Q = np.zeros(m)
    RPM = np.zeros(m)
    eta = np.zeros(m)
    D = np.zeros(m)
    L = np.zeros(m)
    
    endindex = -1
    for i, t in enumerate(ts):
        if SOC[i] < (1-ds):
            endindex = i
            break
        elif V[i] > (4/5)*Vlof: # assume rotation at 4/5 Vlof (BIG BIG APPROX)
            endindex = i
            break
        
        T[i], P[i], Itot[i], RPM[i], Q[i], eta[i] = propulsions.ModelCalcsExternalSOC(V[i], SOC[i], rpm_list, NUMBA_PROP_DATA, CB, ns, Rint, KV, Rm, nmot, I0, ds, takeoffdT, GR)         
        D[i] = 0.5*rho*(V[i]**2)*Sw*CDtoPreR*CDige_oge
        L[i] = 0.5*rho*(V[i]**2)*Sw*CLtoPreR*CLige_oge
        F = mufric*(MGTOW-L[i])
        
        a[i] = (T[i] - D[i] - F)/mass
        
        # PROBLEM: this is derivative based
        V[i+1] = V[i] + a[i]*dt
        x[i+1] = x[i] + V[i]*dt + 0.5*a[i]*(dt**2)
        
        # This is integral based though?
        # SOC[i+1] = (CB*3.6 - trapezoid(Itot[:i], x=ts[:i]))/(CB*3.6)
        SOC[i+1] = SOC[i] - (Itot[i]*dt)/(CB*3.6)
    
    if V[endindex] < (4/5)*Vlof:
        raise ValueError('Aircraft cannot get to rotation velocity!\n\t\t\tDrag, MGTOW, or surface friction must be reduced.')

    # trim arrays
    a = a[:endindex]
    V = V[:endindex]
    x = x[:endindex]
    ts = ts[:endindex]
    T = T[:endindex]
    D = D[:endindex]
    SOC = SOC[:endindex]
    Itot = Itot[:endindex]
    RPM = RPM[:endindex]
    P = P[:endindex]
    Q = Q[:endindex]   
    eta = eta[:endindex]
    
    
    # now simulate post rotation (please don't mention my awful naming conventions)
    
    # updated using the high aoa formulations from Phillips and Hunsaker
    Beta_D = 1 + 0.0361*(CLtoPostR**1.21)/((AR**1.19)*((h/b)**1.51))
    CDige_oge = (1 - delta_D*(np.e**(-4.74*(h/b)**0.814)) - ((h/b)**2)*(np.e**(-3.88*(h/b)**0.758)))*Beta_D
    Beta_L = 1 + 0.269*(CLtoPostR**1.45)
    CLige_oge = (1 + (delta_L*(288*(h/b)**0.787)*(np.e**(-9.14*((h/b)**0.327))))/(AR**0.882))/Beta_L

    ts2 = np.linspace(ts[-1], texpect, m) # start the mission at 0.0 s
    dt2 = ts[2]-ts[1]
    
    x2 = np.zeros(m+1)
    V2 = np.zeros(m+1)
    SOC2 = np.zeros(m+1)

    a2 = np.zeros(m)
    T2 = np.zeros(m)
    P2 = np.zeros(m)
    Itot2 = np.zeros(m)
    Q2 = np.zeros(m)
    RPM2 = np.zeros(m)
    D2 = np.zeros(m)
    L2 = np.zeros(m)
    eta2 = np.zeros(m)
    
    V2[0] = V[-1]      # initial V (m/s)
    SOC2[0] = SOC[-1]  # initial state of charge
    x2[0] = x[-1]
    Itot2[0] = Itot[-1] # also need to do this so SOC doesn't reset to 100!    
    
    endindex = -1
    for i, t in enumerate(ts2):
        if SOC2[i] < (1-ds):
            endindex = i
            
            break
        elif V2[i] > Vlof: # assume rotation at 4/5 Vlof (BIG BIG APPROX)
            endindex = i
            break
        
        T2[i], P2[i], Itot2[i], RPM2[i], Q2[i], eta2[i] = propulsions.ModelCalcsExternalSOC(V2[i], SOC2[i], rpm_list, NUMBA_PROP_DATA, 
                                                 CB, ns, Rint, KV, Rm, nmot, I0, ds, takeoffdT, GR)         
        D2[i] = 0.5*rho*(V2[i]**2)*Sw*CDtoPreR*CDige_oge
        L2[i] = 0.5*rho*(V2[i]**2)*Sw*CLtoPreR*CLige_oge
        F = mufric*(MGTOW-L2[i])
        
        a2[i] = (T2[i] - D2[i] - F)/mass
        
        V2[i+1] = V2[i] + a2[i]*dt2
        x2[i+1] = x2[i] + V2[i]*dt2 + 0.5*a2[i]*(dt2**2)
        SOC2[i+1] = SOC2[i] - (Itot2[i]*dt2)/(CB*3.6)

    if V2[endindex] < Vlof:
        raise ValueError('Aircraft cannot takeoff!\n\t\t\tDrag, MGTOW, or surface friction must be reduced.')

    # trim arrays
    a2 = a2[:endindex]
    V2 = V2[:endindex]
    x2 = x2[:endindex]
    ts2 = ts2[:endindex]
    T2 = T2[:endindex]
    D2 = D2[:endindex]
    SOC2 = SOC2[:endindex]
    Itot2 = Itot2[:endindex]
    RPM2 = RPM2[:endindex]
    P2 = P2[:endindex]
    Q2 = Q2[:endindex]
    eta2 = eta2[:endindex]
        
    # combine (this is so badly done, there MUST be a better way) (could use a list and iterate but whatever)
    a = np.append(a, a2)
    h = np.zeros(a.size) # no altitude gained during ground roll!
    n = np.zeros(a.size) # not calculating load factor for takeoff
    V = np.append(V, V2)
    x = np.append(x, x2)
    ts = np.append(ts, ts2)
    T = np.append(T, T2)
    D = np.append(D, D2)
    SOC = np.append(SOC, SOC2)
    Itot = np.append(Itot, Itot2)
    RPM = np.append(RPM, RPM2)
    P = np.append(P, P2)
    Q = np.append(Q, Q2)
    eta = np.append(eta, eta2)
    
    ############# NOW INITIAL CLIMB ###############
    
    
    
    
    
    
    
    
    
    
    return(a, V, x, h, ts, T, D, n, SOC, Itot, RPM, P, Q, eta)

#%% Cruise
@njit
def Cruise(segment_distance, V_initial, t_initial, SOC_initial, x_initial, h_initial, eta_initial, CL0, CD0, 
           Sw, rho, MGTOW, mass, ds, rpm_list, NUMBA_PROP_DATA, CB, ns, Rint, KV, Rm, nmot, I0, dT, GR, tend = 500, m = 1000):
    '''
    Segment distance in m
    Initial velocity in m/s
    initial time in s
    initial SOC in %/100 
    CL0, CD0 aerodynamic coefs at cruise conditions (usually 0 aoa)
    rho = air density kg/m3
    mass = plane mass in kg
    
    max usable ds is 0.9999 (1.0 breaks the curve!)
    '''
    if segment_distance < 0:
        raise ValueError('Segment Distance < 0')
    
    ts = np.linspace(t_initial, tend, m)
    dt = ts[2]-ts[1]
    
    x = np.zeros(m+1)
    h = np.ones(m+1)*h_initial
    V = np.zeros(m+1)
    a = np.zeros(m)
    T = np.zeros(m)
    P = np.zeros(m)
    Itot = np.zeros(m)
    Q = np.zeros(m)
    eta = np.zeros(m)
    RPM = np.zeros(m)
    D = np.zeros(m)
    n = np.ones(m)
    
    SOC = np.zeros(m+1)
    SOC[0] = SOC_initial #initial state of charge
    V[0] = V_initial #initial V (m/s)
    x[0] = x_initial
    eta[0] = eta_initial
    
    # propulsions.ModelCalcs(self, V[i], t)
    # what if I want to account for changing Itot???
    # realistically I should aim to use RK45 or something 
    # need to handle an error where the input SOC is < 1-ds already!
    # print(SOC_initial)
    # print(SOC[0])
    for i, t in enumerate(ts):
        if SOC[i] < (1-ds):
            endindex = i
            break
        elif x[i]-x_initial > segment_distance:
            endindex = i
            break
        
        # essentially bc CL switches instantaneously between turn and cruise, 
        # for a brief moment the lift created at CL0 is less than MGTOW
        # elif 0.5*rho*(V[i]**2)*Sw*CL0 < MGTOW:
            # print(0.5*rho*(V[i]**2)*Sw*CL0)
            # print(MGTOW)
            # Vstall = np.sqrt((2*MGTOW)/(rho*Sw*CL0))
            # print(Vstall, V[i])
            # raise ValueError('Aircraft too heavy for level flight')
        
        T[i], P[i], Itot[i], RPM[i], Q[i], eta[i] = propulsions.ModelCalcsExternalSOC(V[i], SOC[i], rpm_list, NUMBA_PROP_DATA, 
                                                                              CB, ns, Rint, KV, Rm, nmot, I0, ds, dT, GR) 
        D[i] = 0.5*rho*(V[i]**2)*Sw*CD0
        a[i] = (T[i]-D[i])/mass
        V[i+1] = V[i] + a[i]*dt
        x[i+1] = x[i] + V[i]*dt + 0.5*a[i]*(dt**2)
        SOC[i+1] = SOC[i] - (Itot[i]*dt)/(CB*3.6)
        
    a = a[:endindex]
    V = V[:endindex]
    x = x[:endindex]
    h = h[:endindex]
    ts = ts[:endindex]
    T = T[:endindex]
    D = D[:endindex]
    n = n[:endindex]
    SOC = SOC[:endindex]
    Itot = Itot[:endindex]
    RPM = RPM[:endindex]
    P = P[:endindex]
    Q = Q[:endindex]
    eta = eta[:endindex]
    
    return(a, V, x, h, ts, T, D, n, SOC, Itot, RPM, P, Q, eta)

#%% Climb (essentially cruise but with a theta modifier)
@njit
def Climb(climb_altitude, theta, max_segment_distance, V_initial, t_initial, SOC_initial, x_initial, h_initial, eta_initial, CL0, CD0, 
           Vstall, Sw, rho, MGTOW, mass, ds, rpm_list, NUMBA_PROP_DATA, CB, ns, Rint, KV, Rm, nmot, I0, dT, GR, tend = 500, m = 1000):
    '''
    climb alititude in m
    theta (climb angle) in deg
    h_initial is initial height in m
    
    Assumptions:
        - Start the climb half a meter off the ground (no initial climb modeled)
        - No ground effect corrections implmeented
    
    Should I add altitude to relevant variables? 
    Not really useful here but I think it'll add a lot of versatility, esp with later atmospheric modules
    
    '''
    
    ts = np.linspace(t_initial, tend, m)
    dt = ts[2]-ts[1]
    
    x = np.zeros(m+1)
    V = np.zeros(m+1)
    a = np.zeros(m)
    T = np.zeros(m)
    P = np.zeros(m)
    Itot = np.zeros(m)
    Q = np.zeros(m)
    eta = np.zeros(m)
    RPM = np.zeros(m)
    D = np.zeros(m)
    n = np.zeros(m)
    
    h = np.zeros(m+1) #altitude now
    
    SOC = np.zeros(m+1)
    SOC[0] = SOC_initial #initial state of charge
    V[0] = V_initial #initial V (m/s)
    x[0] = x_initial
    h[0] = h_initial #start half a meter above the ground
    eta[0] = eta_initial
    
    theta = theta*np.pi/180 #convert from deg to rad
    for i, t in enumerate(ts):
        if SOC[i] < (1-ds):
            endindex = i
            break
        elif np.abs(h[i]-h_initial) > climb_altitude:
            endindex = i
            break
        elif x[i]-x_initial > max_segment_distance:
            endindex = i
            break
        
        T[i], P[i], Itot[i], RPM[i], Q[i], eta[i] = propulsions.ModelCalcsExternalSOC(V[i], SOC[i], rpm_list, NUMBA_PROP_DATA, 
                                                                              CB, ns, Rint, KV, Rm, nmot, I0, ds, dT, GR) 
        D[i] = 0.5*rho*(V[i]**2)*Sw*CD0
        # L = 0.5*rho*(V[i]**2)*Sw*CL0
        
        Wuse = mass*9.81*np.sin(theta) # weight at a certain climb angle
        a[i] = (T[i] - D[i] - Wuse)/mass # acceleration along the path
        
        V[i+1] = V[i] + a[i]*dt
        n[i] = np.cos(theta) #n = L/W, in steady climb that's L*cos   #(T[i]/MGTOW)*(L/D[i])

        # if L < MGTOW*np.cos(theta):
        #     print(L)
        #     print(MGTOW)
        #     print(MGTOW*np.cos(theta))
        #     # print((T[i]*np.sin(theta) - D[i]*np.sin(theta) + L*np.cos(theta) - MGTOW)/mass) # ACTUAL acceleration in the y direction
        #     # print(a[i])
        #     raise ValueError('Aircraft cannot climb! Try reducing climb angle or MGTOW.')
        
        x[i+1] = x[i] + V[i]*dt*np.cos(theta) + 0.5*a[i]*(dt**2)*np.cos(theta)
        h[i+1] = h[i] + V[i]*dt*np.sin(theta) + 0.5*a[i]*(dt**2)*np.cos(theta) # does not account for the scenario where L < MGTOW*cos(theta)!
        
        
        # I'm not getting this to work right so we'll keep it simple for now. I think the problem in the descent is that L >> T and D for small aoa
        # IRL you have to reduce throttle to descend right???
        # ay = (T[i]*np.sin(theta) - D[i]*np.sin(theta) + L*np.cos(theta))/mass - 9.81 
        # # if theta < 0:
        # #     print(T[i]*np.sin(theta))
        # #     print(D[i]*np.sin(theta))
        # #     print(L*np.cos(theta))
        # #     print((T[i]*np.sin(theta) - D[i]*np.sin(theta) + L*np.cos(theta))/mass)
        # h[i+1] = h[i] + V[i]*dt*np.sin(theta) + 0.5*ay*(dt**2)
        
        # using specific excess power instead (assumes theta ~= sin(theta), small angle approx)
        # dh_dt = (V[i]/MGTOW)*(T[i]-D[i])
        # h[i + 1] = h[i] + dh_dt*dt
        
        if theta > 0 and h[i+1] <= h[i]:
            raise ValueError('Aircraft cannot climb! Try reducing climb angle or MGTOW.')
        
        # determining acceleration change
        SOC[i+1] = SOC[i] - (Itot[i]*dt)/(CB*3.6)
    
    a = a[:endindex]
    V = V[:endindex]
    x = x[:endindex]
    h = h[:endindex]
    ts = ts[:endindex]
    T = T[:endindex]
    D = D[:endindex]
    n = n[:endindex]
    SOC = SOC[:endindex]
    Itot = Itot[:endindex]
    RPM = RPM[:endindex]
    P = P[:endindex]
    Q = Q[:endindex]
    eta = eta[:endindex]

    return(a, V, x, h, ts, T, D, n, SOC, Itot, RPM, P, Q, eta)


#%% Turn (with bank angle implementation)
def Turn(segment_degrees, nmaxlimit, V_initial, t_initial, SOC_initial, x_initial, h_initial, eta_initial, CLturn, CDturn, 
         Vstall, Sw, rho, MGTOW, mass, ds, rpm_list, NUMBA_PROP_DATA, CB, ns, Rint, KV, Rm, nmot, I0, dT, GR, tend = 15, m = 1000):
    '''
    Assumptions: 
        - Sustained level turns
        - CL/CD at 10 deg aoa
        - g = 9.81 m/s^2
    '''
    g = MGTOW/mass
    
    ts = np.linspace(t_initial, tend, m)
    dt = ts[2]-ts[1]
    
    x = np.zeros(m+1)
    V = np.zeros(m+1)
    a = np.zeros(m)
    T = np.zeros(m)
    P = np.zeros(m)
    Itot = np.zeros(m)
    Q = np.zeros(m)
    eta = np.zeros(m)
    RPM = np.zeros(m)
    D = np.zeros(m)
    h = np.ones(m+1)*h_initial #altitude
    
    # add in load factor to all or no? I say yes
    n = np.zeros(m+1)
    n[0] = 1.0 # from level flight
    deg = np.zeros(m+1)
        
    # specific quantities for turns
    L = np.zeros(m)
    turnrate = np.zeros(m)
    turnradius = np.zeros(m)
    
    SOC = np.zeros(m+1)
    SOC[0] = SOC_initial #initial state of charge
    V[0] = V_initial #initial V (m/s)
    x[0] = x_initial
    h[0] = h_initial #start half a meter above the ground
    n[0] = 1.0 #start from cruise
    eta[0] = eta_initial
    
    for i, t in enumerate(ts):
        if SOC[i] < (1-ds):
            endindex = i
            break
        elif deg[i] > segment_degrees:
            endindex = i
            break
            
        T[i], P[i], Itot[i], RPM[i], Q[i], eta[i] = propulsions.ModelCalcsExternalSOC(V[i], SOC[i], rpm_list, NUMBA_PROP_DATA, 
                                                                              CB, ns, Rint, KV, Rm, nmot, I0, ds, dT, GR) 
        D[i] = 0.5*rho*(V[i]**2)*Sw*CDturn
        L[i] = 0.5*rho*(V[i]**2)*Sw*CLturn
        
        # T/W*L/D assumes the aircraft is not allowed to slow down in turn or lose altitude!! (yet the aircraft DOES lose velocity in the sim)
        # n = L/W for instantaneous (but then how do we model the slowdown or loss in altitude?)
        # how about I just use instantaneous but I also model loss of altitude.
        n[i] = min((T[i]/MGTOW)*(L[i]/D[i]), nmaxlimit) # load factor for sustained turn, pin to nmax if there's a limit
        n[i] = max(n[i], 1.0)
        
        
        ############## NOT CONCLUSIVE WHAT TO DO HERE: REVIEW LATER ##############
        # n[i] = max(L[i]/MGTOW, 1.0) # this essentially forces the aircraft to accelerate if the turn would result in n < 1.0 (i.e. turning directly after a climb)
        # n[i] = min(n[i], nmaxlimit) # cap n at the structural limit!
        
        # Irl you roll to some bank angle which determines your load factor n = 1/cos(phi)
        # but I would like this to work without having to specify bank angle 
        # how do you find the altitude loss without that tho?
        
        # if n[i] < 1:
        #     print(T[i]/MGTOW)
        #     print(L[i]/MGTOW)
        #     print((T[i]/MGTOW)*(L[i]/D[i]))
        #     raise ValueError(f'Load factor of {n[i]:.3f} in sustained turn\n\t\t\tAircraft cannot maintain altitude while turning!')
        
        turnrate[i] = ((g*np.sqrt(n[i]**2 - 1))/V[i])*(180/np.pi) # for some scenario n < 1??
        turnradius[i] = (V[i]**2)/(g*np.sqrt(n[i]**2-1))
        deg[i+1] = deg[i] + turnrate[i]*dt
        
        
        a[i] = (T[i]-D[i])/mass
        V[i+1] = V[i] + a[i]*dt
        x[i+1] = x[i] + V[i]*dt + 0.5*a[i]*(dt**2)
                
        # assume level altitude (for now, accurately calculating climbing turns is my dream but I think it will have to wait for 6DOF)
        # h[i+1] = h[i] + V[i]*dt + 0.5*a[i]*(dt**2)
        
        # determining acceleration change
        SOC[i+1] = SOC[i] - (Itot[i]*dt)/(CB*3.6)
        
    a = a[:endindex]
    V = V[:endindex]
    x = x[:endindex]
    h = h[:endindex]
    ts = ts[:endindex]
    T = T[:endindex]
    D = D[:endindex]
    n = n[:endindex]
    SOC = SOC[:endindex]
    Itot = Itot[:endindex]
    RPM = RPM[:endindex]
    P = P[:endindex]
    Q = Q[:endindex]
    eta = eta[:endindex]
        
    return(a, V, x, h, ts, T, D, n, SOC, Itot, RPM, P, Q, eta) # one little thing up here can mess up everything loll (took 2 hrs at 2 am to debug)
    
    
    
    
    
    
    
    
    
    
    


# -*- coding: utf-8 -*-
"""
OLD FILE; IGNORE

Created on Tue Mar 17 14:49:22 2026

Test for hicken

@author: sammy
"""

import common as ud
import matplotlib.pyplot as plt
import numpy as np
import scipy.optimize as sop

design = ud.PointDesign()
# design.OpenMotorData()
design.Motor('V8110-170', 1)
design.Battery('Gaoneng_8S_3300', 0.85)
design.Prop('22x12E')

design.Motor('C-4130/20', 1)
design.Battery('Gaoneng_8S_3300', 0.85)
design.Prop('22x12E')
# design.PointResult(Uinf = 28, h = 10, SOC = 1.0, dT = 0.8, verbose = True)


Uinf = np.linspace(0, 30, 50)
propQs, arr = design.LinePlot(Uinf = Uinf, h = 10, SOC = 0.6, dT = 0.4, verbose = False)

Tcoefs = np.polyfit(arr, propQs[:, 0], 4)
Tfit = np.poly1d(Tcoefs)
rho = 1.224 
CdSw = 0.12*0.87
def T_func(V):
    return(Tfit(V) - 0.5*rho*(V**2)*CdSw)

Vmatch = sop.root(T_func, (Uinf[-1]-Uinf[0])/2)['x'][0]

# T, D plot
fig, ax = plt.subplots(figsize = (6, 4), dpi = 500)
                
propQnames = ['Total Thrust (N)', 
              'Total Torque (Nm)', 
              'RPM', 
              'Drive Efficiency', 
              'Propeller Efficiency', 
              'Gearing Efficiency', 
              'Motor Efficiency', 
              'ESC Efficiency', 
              'Battery Efficiency', 
              'Mech. Power Out of 1 Motor (W)', 
              'Elec. Power Into 1 Motor (W)', 
              'Elec. Power Into 1 ESC (W)', 
              'Waste Power in 1 Motor (W)', 
              'Waste Power in 1 ESC (W)',
              'Waste Power in 1 Battery (W)',
              'Current in 1 Motor (A)', 
              'Current in 1 ESC (A)', 
              'Current in Battery (A)',
              'Voltage in 1 Motor (V)', 
              'Voltage in 1 ESC (V)', 
              'Battery Voltage (V)', 
              'Voltage Per Cell (V)', 
              'State of Charge']
ax.plot(arr, propQs[:, 0], label = 'Thrust output')
ax.plot(arr, Tfit(arr), '--', label = 'Thrust fit')
ax.plot(arr, 0.5*rho*(arr**2)*CdSw, label = 'Drag')
ax.minorticks_on()
plt.ylabel(f'{propQnames[0]}') # need to add a way to get units in correctly
plt.title('Velocity sweep for 300m altitude, 70% SOC, 100% throttle')
plt.xlabel('Velocity (m/s)')
plt.legend()
plt.show()

#%% Checking drive efficiency
fig, ax = plt.subplots(figsize = (6, 4), dpi = 500)
                
Pb = propQs[:, 11] # Pin to 1 ESC
# 14 current, 17 voltage
Pb_alt = propQs[:, 17]*propQs[:, 20]

# check that Pb = Pb_alt
# print(Pb[5], Pb_alt[5])
# True!

# then check (T*Uinf)/(Pb+Plostinm) = eta_drive
# print(propQs[:, 0]*Uinf/Pb) # without P_wastebat
# print(propQs[:, 3])
# they're nearly identical as expected (differing in the 4th or 5th decimal place)

# ax.plot(arr, propQs[:, 0], label = 'Thrust output')
# ax.plot(arr, Tfit(arr), '--', label = 'Thrust fit')
# ax.plot(arr, 0.5*rho*(arr**2)*CdSw, label = 'Drag')
# ax.minorticks_on()
# plt.ylabel(f'{propQnames[0]}') # need to add a way to get units in correctly
# plt.title('Velocity sweep for 300m altitude, 70% SOC, 100% throttle')
# plt.xlabel('Velocity (m/s)')
# plt.legend()
# plt.show()


#%% Other quantities with velocity match plotted 
Vs = arr 
for interestindex in [0, 2, 3, 4, 6, 12, 13, 14]:
    fig, ax = plt.subplots(figsize = (6, 4), dpi = 500)
    ax.plot(arr, propQs[:, interestindex])
    ax.axvline(Vmatch, linestyle = '--', color = 'k', label = 'T = D')
    ax.grid()
    ax.minorticks_on()
    plt.ylabel(f'{propQnames[interestindex]}') # need to add a way to get units in correctly
    plt.title('Velocity sweep for 300m altitude, 70% SOC, 100% throttle')
    plt.xlabel('Velocity (m/s)')

